﻿module Md2Read

open System
open System.Collections.Generic
open System.Threading
open System.Drawing
open System.IO
open OpenTK
open OpenTK.Graphics
open OpenTK.Graphics.OpenGL
open Common
open CgNet
open CgNet.GL
open System.Runtime.InteropServices

type ArrayList<'T> = System.Collections.Generic.List<'T>

type Md2Header =
    struct
        val magic:           int         //MD2文件标示   
        val version:         int         //MD2版本
        val skinWidth:       int         //纹理宽度
        val skinHeight:      int         //纹理长度
        val frameSize:       int         //桢的大小
        val numSkins:        int         //
        val numVertices:     int         //多少个顶点(每桢数量相同，数据不同)
        val numTexCoords:    int         //多少个纹理顶点(所有桢共用)
        val numTriangles:    int         //每桢由多少个三角形组成，所有桢是一样的
        val numGlCommands:   int         //用VBO直接放弃
        val numFrames:       int         //多少桢
        val offsetSkins:     int         //
        val offsetTexCoords: int         //从那开始读纹理数据
        val offsetTriangles: int         //从那开始读三角形
        val offsetFrames:    int         //从那开始读桢数据
        val offsetGlCommands:int         //无用
        val offsetEnd:       int         //可以用来检查
    end    

type Md2Frame(md2Model: Md2Model,count:int) =
    let mutable points = Array2D.create 0 0 0.f
    let mutable vbo = 0
    member val Vectexs = Array.create count Vector3.Zero
    member val Name = ""
    member this.VBO with get() = vbo
    member this.Faces with get() : ArrayList<int[]*int[]> = md2Model.Faces
    member this.ElementCount with get() = md2Model.ElementCount
    member this.DataArray 
        with get() = 
            if points.Length = 0 then this.CreateData()
            points
    //MD2中不变的是面的面数.面里的顶点根据桢里保存的不同而不同,而面用的纹理是用的同一数据
    member this.CreateData() =
        let normals = Array.create this.Vectexs.Length (0.f,Vector3.Zero)
        //遍历第一次,生成面法线,记录对应点的共面,共法线信息
        this.Faces.ForEach(fun p ->
            let vi = fst p
            let p1 = this.Vectexs.[vi.[1]] - this.Vectexs.[vi.[0]]
            let p2 = this.Vectexs.[vi.[2]] - this.Vectexs.[vi.[0]]
            let normal = -Vector3.Cross(p1,p2)
            vi |> Array.iter(fun v ->
                let mutable ind,n = normals.[v]
                n <- n + normal
                normals.[v] <- (ind+1.f,n)
                )
            )
        //平均点的法线信息并且组装N3fV3f
        points <- Array2D.init this.ElementCount 6 (fun i j ->
            //当前面,当前面的第几个点
            let m,n = i/3,i%3
            let vi = fst this.Faces.[m]
            match j with
            | 0|1|2 -> 
                let vn = snd normals.[vi.[n]]/fst normals.[vi.[n]]
                if j = 0 then vn.X elif j = 1 then vn.Y else vn.Z  
            | 3|4|5 -> 
                let p = this.Vectexs.[vi.[n]]
                if j = 3 then p.X elif j = 4 then p.Y else p.Z
            )        
    member this.CreateVBO() =
        if vbo = 0 then
            vbo <- GL.GenBuffers(1)
            GL.BindBuffer(BufferTarget.ArrayBuffer,vbo)
            GL.BufferData(BufferTarget.ArrayBuffer,IntPtr (4 * this.ElementCount * 6),this.DataArray,BufferUsageHint.StaticDraw)

and Md2Model(fileName:string,?texureName:string) =
    inherit ModelCommon()
    let mutable vbo,ebo = 0,0
    member val Header = Md2Header() with get,set
    member val Faces = new ArrayList<int[]*int[]>()
    member val TexCoords = new ArrayList<float32*float32>()
    member val Frames = new ArrayList<Md2Frame>()
    member val texID = 0 with get,set  
    member val CurrentFrame = 0.f with get,set  
    member this.ElementCount with get() = this.Faces.Count * 3
    member this.TotalFrames with get() = this.Frames.Count
    member this.LoadModel() =  
        //加载纹理
        if texureName.IsSome then
            let dict = Path.GetDirectoryName(fileName)
            this.texID <- TexTure.Load(Path.Combine(dict,texureName.Value))
        //加载MD2程序
        let file = new FileStream(fileName,FileMode.Open, FileAccess.Read)
        let binary = new BinaryReader(file)
        let size = Marshal.SizeOf(this.Header)
        let mutable bytes = Array.create size 0uy
        file.Read(bytes, 0, bytes.Length) |> ignore
        let allocIntPtr = Marshal.AllocHGlobal(size)
        Marshal.Copy(bytes,0,allocIntPtr,size)
        this.Header <- Marshal.PtrToStructure(allocIntPtr,typeof<Md2Header>) :?> Md2Header 
        //读取纹理数据         
        file.Seek(int64 this.Header.offsetTexCoords,SeekOrigin.Begin)|> ignore
        let mTexCoords = Array.init this.Header.numTexCoords (fun p -> 
            float32 (binary.ReadInt16())/float32 this.Header.skinWidth,
            float32 (binary.ReadInt16())/float32 this.Header.skinWidth
            )
        this.TexCoords.AddRange(mTexCoords)
        //读取面数(顶点索引与纹理索引)
        file.Seek(int64 this.Header.offsetTriangles,SeekOrigin.Begin)|> ignore
        let mtriangles = Array.init this.Header.numTriangles (fun p ->
            [|int (binary.ReadInt16());int (binary.ReadInt16());int (binary.ReadInt16())|],
            [|int (binary.ReadInt16());int (binary.ReadInt16());int (binary.ReadInt16())|]
            )
        this.Faces.AddRange(mtriangles)
        //读取所有桢
        file.Seek(int64 this.Header.offsetFrames,SeekOrigin.Begin)|> ignore
        let frames = Array.init this.Header.numFrames (fun p -> 
            let frame = Md2Frame(this,this.Header.numVertices)
            let scale = Vector3(binary.ReadSingle(),binary.ReadSingle(),binary.ReadSingle())
            let translate = Vector3(binary.ReadSingle(),binary.ReadSingle(),binary.ReadSingle())
            let name = binary.ReadChars(16)
            //这桢的所有点
            let vectexs = Array.init this.Header.numVertices (fun t -> 
                let mvertex = [|binary.ReadByte();binary.ReadByte();binary.ReadByte()|]
                let mlightNormalIndex = binary.ReadByte()
                mvertex,mlightNormalIndex
                )
            //桢上的点精确化
            vectexs |> Array.iteri(fun i v ->
                frame.Vectexs.[i].X <- float32 (fst v).[0] * scale.X + translate.X
                frame.Vectexs.[i].Y <- float32 (fst v).[2] * scale.Z + translate.Z
                frame.Vectexs.[i].Z <- float32 (fst v).[1] * -scale.Y - translate.Y
                )
            frame
            )
        this.Frames.AddRange(frames)
        //生成正确的数据
        binary.Close()
        file.Close()
    member this.FrameStep 
        with get() =
           let currentFrame = int (Math.Floor(float this.CurrentFrame))   
           let step = this.CurrentFrame - float32 currentFrame 
           currentFrame,step
    member this.CreateEBO() =
        let len = this.ElementCount - 1
        let eboData = [|0..len|]
        ebo <- GL.GenBuffers(1)
        GL.BindBuffer(BufferTarget.ElementArrayBuffer,ebo)
        GL.BufferData(BufferTarget.ElementArrayBuffer,IntPtr (4 * this.ElementCount),eboData,BufferUsageHint.StaticDraw)
    member this.CreateVBO() =
        let texCoords = Array2D.init this.ElementCount 2 (fun i j ->
            //当前面,当前面的第几个点
            let m,n = i/3,i%3
            let ti = snd this.Faces.[m]
            let u,v = this.TexCoords.[ti.[n]]
            if j = 0 then u else v
            )  
        vbo <- GL.GenBuffers(1)
        GL.BindBuffer(BufferTarget.ArrayBuffer,vbo)
        GL.BufferData(BufferTarget.ArrayBuffer,IntPtr (4 * 2 * this.ElementCount),texCoords,BufferUsageHint.StaticDraw)
    member this.Render() =
        if vbo = 0 then this.CreateVBO()
        if ebo = 0 then this.CreateEBO()  
        if this.CurrentFrame >= float32 this.TotalFrames - 1.f then this.CurrentFrame <- 0.f
        let currentFrame = this.Frames.[fst this.FrameStep]
        let nextFrame = this.Frames.[fst this.FrameStep + 1]
        currentFrame.CreateVBO()
        nextFrame.CreateVBO() 
        //当前桢的法线与顶点
        GL.BindBuffer(BufferTarget.ArrayBuffer,currentFrame.VBO)
        GL.VertexAttribPointer(1,3,VertexAttribPointerType.Float,false,24,IntPtr 12)
        GL.EnableVertexAttribArray(1)       
        GL.VertexAttribPointer(2,3,VertexAttribPointerType.Float,false,24,IntPtr.Zero)
        GL.EnableVertexAttribArray(2)
        //如果有纹理        
        if this.texID > 0 && vbo > 0 then    
            GL.BindBuffer(BufferTarget.ArrayBuffer,vbo)   
            GL.VertexAttribPointer(8,2,VertexAttribPointerType.Float,false,0,IntPtr.Zero)
            GL.EnableVertexAttribArray(8)
        //下一桢的法线与顶点存放在Texture1与Texture2
        GL.BindBuffer(BufferTarget.ArrayBuffer,nextFrame.VBO)   
        GL.VertexAttribPointer(9,3,VertexAttribPointerType.Float,false,24,IntPtr 12)
        GL.EnableVertexAttribArray(9)
        GL.VertexAttribPointer(10,3,VertexAttribPointerType.Float,false,24,IntPtr.Zero)
        GL.EnableVertexAttribArray(10)
        //绘画
        GL.BindBuffer(BufferTarget.ElementArrayBuffer,ebo)
        GL.DrawElements(BeginMode.Triangles,this.ElementCount,DrawElementsType.UnsignedInt,IntPtr.Zero)


