﻿module Md5File

open System
open System.Collections.Generic
open System.Threading
open System.Drawing
open System.IO
open OpenTK
open OpenTK.Graphics
open OpenTK.Graphics.OpenGL
open Common
open CgNet
open CgNet.GL
open System.Runtime.InteropServices

type ArrayList<'T> = System.Collections.Generic.List<'T>
let filtLine (line:string) = not (String.IsNullOrEmpty(line)) && line <> "(" && line <> ")" 
let getLineData (line:string) = line.Split(' ','\t','\"') |> Array.filter filtLine
let getFloat str = snd (System.Single.TryParse(str))
let getInt str = snd (System.Int32.TryParse(str))
let getw x y z = 1.0f - x*x - y*y - z*z |> fun p -> if p < 0.f then 0.f else float32 (-Math.Sqrt(float p))
//--------mesh里用的结构
type Md5Joint() = 
    member val Name = "" with get,set
    member val Index = -2 with get,set
    //位移
    member val Position = Vector3.Zero with get,set
    //旋转
    member val Quat = Quaternion.Identity  with get,set       
    member val ParentIndex = -2 with get,set    
    member this.SetValue(line:string,ind:int) =
        let ls = getLineData line// line.Split(' ','\t') |> Array.filter (fun p -> not (String.IsNullOrEmpty(p)))//&& p <> @"\t" && p.Length > 0)
        let pos,quat =
            let gf str = snd (System.Single.TryParse(str))
            let x,y,z,a,b,c = (gf ls.[2]),(gf ls.[3]),(gf ls.[4]),(gf ls.[5]),(gf ls.[6]),(gf ls.[7])
            let d = getw a b c
            Vector3(x,y,z),Quaternion(a,b,c,d)
        this.Name <- ls.[0]
        this.ParentIndex <- snd (System.Int32.TryParse(ls.[1]))
        this.Position <- pos
        this.Quat <- quat
        this

//Vert包含纹理坐标，以及关联的权重，根据权重求顶点
type Md5Vert() =
    member val Index = -1 with get,set
    member val Texcoord = Vector2.Zero with get,set
    member val WeightStart = 0 with get,set
    member val WeightCount = 0 with get,set
    //经过权重求顶点实际位置
    member val Position = Vector3.Zero with get,set 
    member val Normal = Vector3.Zero with get,set       
    member this.SetValue(line:string) =
        let ls =getLineData line// line.Split(' ','\t') |> Array.filter (fun p -> not (String.IsNullOrEmpty(p)))
        let gf str = snd (System.Single.TryParse(str))
        this.Index <- snd (System.Int32.TryParse(ls.[1]))
        this.Texcoord <- Vector2(gf ls.[2],gf ls.[3])
        this.WeightStart <- snd (System.Int32.TryParse(ls.[4]))
        this.WeightCount <- snd (System.Int32.TryParse(ls.[5]))
        this
    member this.DataArray with get() =[| this.Texcoord.X;this.Texcoord.Y;this.Position.X;this.Position.Y;this.Position.Z|]

//三角形(包含Vert的索引)
type Md5Tri() =
    member val Index = -1 with get,set
    member val VertorIndexs = Array.create 3 0 with get,set
    member this.SetValue(line:string) =
        let ls = getLineData line // line.Split(' ','\t') |> Array.filter (fun p -> not (String.IsNullOrEmpty(p)))
        let gi str = snd (System.Int32.TryParse(str))
        this.Index <- gi ls.[1]
        this.VertorIndexs <- [|gi ls.[2];gi ls.[3];gi ls.[4]|]
        this

//权重，(权重顶点用于计算Md5Vert,权重的JointIndex用于得到对应joint的四元数）
type Md5Weight() =
    member val Index = -1 with get,set
    member val JointIndex = -1 with get,set
    member val Bias = 0.f with get,set
    member val Position = Vector3.Zero with get,set
    member this.SetValue(line:string) =
        let ls = line.Split(' ','\t') |> Array.filter (fun p -> not (String.IsNullOrEmpty(p)))
        let gf str = snd (System.Single.TryParse(str))
        this.Index <- snd (System.Int32.TryParse(ls.[1]))
        this.JointIndex <- snd (System.Int32.TryParse(ls.[2]))
        this.Bias <- gf ls.[3]
        this.Position <- Vector3(gf ls.[5],gf ls.[6],gf ls.[7])
        this

type Md5Mesh() =
    let mutable vbo,ebo = 0,0
    member val TexID = 0 with get,set
    member val ShaderPath = "" with get,set
    member val Verts = ArrayList<Md5Vert>() with get,set
    member val Faces = ArrayList<Md5Tri>() with get,set
    member val Weights = ArrayList<Md5Weight>() with get,set
    member this.ElementCount with get() = this.Faces.Count * 3
    member this.CreateVBO() =
        let vboData = Array2D.init this.ElementCount 5 (fun i j ->
            let a,b = i/3,i%3
            this.Verts.[this.Faces.[a].VertorIndexs.[b]].DataArray.[j]
            )
        vbo <- GL.GenBuffers(1)
        GL.BindBuffer(BufferTarget.ArrayBuffer,vbo)
        GL.BufferData(BufferTarget.ArrayBuffer,IntPtr (4 *this.ElementCount*5),vboData,BufferUsageHint.StaticDraw)
        let len = this.ElementCount - 1
        let eboData = [|0..len|]
        ebo <- GL.GenBuffers(1)
        GL.BindBuffer(BufferTarget.ElementArrayBuffer,ebo)
        GL.BufferData(BufferTarget.ElementArrayBuffer,IntPtr (4 * this.ElementCount),eboData,BufferUsageHint.StaticDraw)    
        if File.Exists this.ShaderPath then
            this.TexID <- TexTure.Load(this.ShaderPath)
    member this.DrawVBO() = 
        if vbo >0 && ebo >0 then
            GL.BindBuffer(BufferTarget.ArrayBuffer,vbo)
            GL.BindBuffer(BufferTarget.ElementArrayBuffer,ebo)
            if this.TexID > 0 then
                GL.Enable(EnableCap.Texture2D)
                GL.BindTexture(TextureTarget.Texture2D,this.TexID)
            GL.InterleavedArrays(InterleavedArrayFormat.T2fV3f,0,IntPtr.Zero)
            GL.DrawElements(BeginMode.Triangles,this.ElementCount,DrawElementsType.UnsignedInt,IntPtr.Zero)
            GL.Disable(EnableCap.Texture2D)        
    member this.Render() =
        let vboData = Array2D.init this.ElementCount 5 (fun i j ->
            let a,b = i/3,i%3
            this.Verts.[this.Faces.[a].VertorIndexs.[b]].DataArray.[j]
            )
        GL.InterleavedArrays(InterleavedArrayFormat.T2fV3f,0,vboData)
        if this.TexID > 0 then
            GL.Enable(EnableCap.Texture2D)
            GL.BindTexture(TextureTarget.Texture2D,this.TexID)       
        GL.DrawElements(BeginMode.Triangles,this.ElementCount,DrawElementsType.UnsignedInt,[|0..this.ElementCount - 1|])
        
//---------anim要用的结构
type Md5JointInfo() =
    member val Name = "" with get,set
    member val Index = -2 with get,set
    member val ParentIndex = -2 with get,set    
    member val Flags = 0 with get,set
    member val StartIndex = 0 with get,set
    member this.SetValue(line:string,ind:int) =
        let ls = getLineData line
        this.Name <- ls.[0]
        this.ParentIndex <- getInt ls.[1]
        this.Index <- ind
        this.Flags <- getInt ls.[2]
        this.StartIndex <- getInt ls.[3]
        this

type Md5BaseFrame() =
    member val Index = -2 with get,set
    //位移
    member val Positions = ArrayList<Vector3>() with get,set
    //旋转
    member val Quats = ArrayList<Quaternion>() with get,set  
    member this.SetValue(line:string) =
        let ls = getLineData line
        let pos,quat =
            let gf str = snd (System.Single.TryParse(str))
            let x,y,z,a,b,c = (gf ls.[0]),(gf ls.[1]),(gf ls.[2]),(gf ls.[3]),(gf ls.[4]),(gf ls.[5])
            let d = getw a b c
            Vector3(x,y,z),Quaternion(a,b,c,d)
        this.Positions.Add(pos)
        this.Quats.Add(quat)

type Md5Frame() =
    member val Index = -2 with get,set
    member val Points = ArrayList<float32>() with get,set
    member this.SetValue(line:string) =
        let ls = getLineData line
        let ds = ls |> Array.map (fun p -> getFloat p)
        this.Points.AddRange(ds)

//桢动画计算得出如下内容
type Md5SkeletonJoin() =
    member val ParentIndex = -2 with get,set
    //位移
    member val Position = Vector3.Zero with get,set
    //旋转
    member val Quat = Quaternion.Identity with get,set  
type Md5FrameSkeleton = ArrayList<Md5SkeletonJoin>
            
type Md5Animation() =
    let mutable currentTime = 0.f
    member val FrameRate = 24.f with get,set    
    //JointInfos集合对象的索引就是本身在文件中的位置，他们本身的顺序就是正序加1
    member val JointInfos = ArrayList<Md5JointInfo>() with get,set
    member val Bounds = ArrayList<Vector3*Vector3>() with get,set
    member val Frames = ArrayList<Md5Frame>() with get,set
    member val BaseFrame = Md5BaseFrame() with get,set    
    member val FrameSkeletonList = ArrayList<Md5FrameSkeleton>() with get,set
    //二个目标,一是转化Frames里的数据成对应一桢的所有骨骼节点信息
    //二是把所有骨骼节点在父骨骼坐标系中的位置转化成模型坐标系
    member this.CreateFrameSkeleton() =
        for frame in this.Frames do
            let md5FrameSkeleton = Md5FrameSkeleton()
            for jointInfo in this.JointInfos do
                //skeleton的顺序因为JointInfos的特殊性，也是正序加1
                let skeleton = Md5SkeletonJoin()
                md5FrameSkeleton.Add(skeleton)
                skeleton.ParentIndex <- jointInfo.ParentIndex
                let position = this.BaseFrame.Positions.[jointInfo.Index]
                let quat = this.BaseFrame.Quats.[jointInfo.Index]
                let setFlags index =
                    let flag = int (Math.Pow(float 2,float index))
                    if jointInfo.Flags &&& flag = flag then frame.Points.[jointInfo.StartIndex + index] 
                    else 
                        match index with 
                        | 0 -> position.X                    
                        | 1 -> position.Y
                        | 2 -> position.Z
                        | 3 -> quat.X
                        | 4 -> quat.Y
                        | 5 -> quat.Z
                        | _ -> quat.W
                let x,y,z,a,b,c = setFlags 0,setFlags 1,setFlags 2,setFlags 3,setFlags 4,setFlags 5
                let w = getw a b c
                //currentPos,currentQuat都是针对父骨骼来的坐标，要转化得到模型坐标
                let currentPos,currentQuat=Vector3(x,y,z),Quaternion(a,b,c,w)
                skeleton.Position <- currentPos
                skeleton.Quat <- currentQuat
                if skeleton.ParentIndex >= 0 then
                    let parentSkeleton = md5FrameSkeleton.[skeleton.ParentIndex]
                    //先得到currentPos经过父骨骼四元数旋转后的值
                    let pos = Vector3.Transform(currentPos,parentSkeleton.Quat)
                    //模型坐标示下的点
                    skeleton.Position <- parentSkeleton.Position + pos
                    //模型坐标系下的四元数
                    skeleton.Quat <- Quaternion.Normalize(parentSkeleton.Quat * skeleton.Quat)
            this.FrameSkeletonList.Add(md5FrameSkeleton) 
    member this.CurrentTime 
        with get() =
            if currentTime > float32 this.Frames.Count/this.FrameRate then currentTime <- 0.f
            currentTime
        and set value = currentTime <- value
    member this.GetCurrentFrameSkeleton() =
        //得到当前的时间所在动画循环的位置
        let current = this.CurrentTime * this.FrameRate
        //得到所在位置的当前桢索引,与运动到下一桢的位置
        let currentFrame,currentStep= int (Math.Floor(float current)),current - float32 (Math.Floor(float current))
        //得到下一桢索引,到桢尾就从头开始
        let nextFrame = if currentFrame < this.Frames.Count - 1 then currentFrame + 1 else 0
        //得到当前桢,下一桢具体信息
        let currentSkeleton,nexSkeleton = this.FrameSkeletonList.[currentFrame],this.FrameSkeletonList.[nextFrame]
        let joints = ArrayList<Md5Joint>()
        for i in [|0 .. this.JointInfos.Count - 1|] do
            //根据当前桢位置求得对应四元数与顶点的插值
            let lerpPosition = Vector3.Lerp(currentSkeleton.[i].Position,nexSkeleton.[i].Position,currentStep)
            let slerpQuat = Quaternion.Slerp(currentSkeleton.[i].Quat,nexSkeleton.[i].Quat,currentStep)
            joints.Add(Md5Joint(Index = i,Position = lerpPosition,Quat = slerpQuat))
        joints

type Md5Model(fileName:string,?animName:string) =
    inherit ModelCommon()
    let mutable vbo,eb0 = 0,0
    member val Joints = ArrayList<Md5Joint>() with get,set
    member val Meshs = ArrayList<Md5Mesh>() with get,set
    member val Animation = Md5Animation() with get,set
    member this.CurrentTime with get() = this.Animation.CurrentTime and set value = this.Animation.CurrentTime <- value
    member this.LoadModel() =  
        let (|StartsWith|) suffix (s:string) = s.TrimStart(' ','\t').StartsWith(suffix,StringComparison.OrdinalIgnoreCase)
        let file = new StreamReader(fileName)
        while not file.EndOfStream  do
            let str = file.ReadLine()
            match str with
            | StartsWith "joints" true  -> 
                let mutable isJoint = true
                while isJoint  do
                    let joint = file.ReadLine()
                    isJoint <- not (joint.Contains("}"))
                    if isJoint then this.Joints.Add(Md5Joint().SetValue(joint,this.Joints.Count))                    
            | StartsWith "mesh" true ->
                let mutable isMesh = true
                let md5mesh = Md5Mesh()
                this.Meshs.Add(md5mesh)
                while isMesh do
                    let mesh = file.ReadLine()
                    match mesh with
                    | StartsWith "shader" true  -> 
                        let dict = Path.GetDirectoryName(fileName)
                        let fileName = (getLineData mesh).[1]// mesh.Split(' ','\t') |> Array.filter (fun p -> not (String.IsNullOrEmpty(p))) |> fun p -> p.[1].Trim('\"')
                        md5mesh.ShaderPath <- Path.Combine(dict,fileName)
                    | StartsWith "vert" true  -> 
                        md5mesh.Verts.Add(Md5Vert().SetValue(mesh))
                    | StartsWith "tri" true  -> 
                        md5mesh.Faces.Add(Md5Tri().SetValue(mesh))
                    | StartsWith "weight" true  -> 
                        md5mesh.Weights.Add(Md5Weight().SetValue(mesh))
                    | StartsWith "}" true  -> 
                        isMesh <- false
                    | _ -> printfn "%s" ("---------"+str) 
            | _ -> printfn "%s" str        
        file.Close()
        if animName.IsSome then
            let path = Path.Combine(Path.GetDirectoryName(fileName),animName.Value)
            let animFile = new StreamReader(path)
            while not animFile.EndOfStream  do
                let str = animFile.ReadLine()
                match str with
                | StartsWith "frameRate" true  -> 
                    this.Animation.FrameRate <- getFloat (getLineData str).[1]  
                | StartsWith "hierarchy" true ->
                    let mutable isJoinHierarchy = true 
                    while isJoinHierarchy  do
                        let joint = animFile.ReadLine()
                        isJoinHierarchy <- not (joint.Contains("}"))
                        if isJoinHierarchy then this.Animation.JointInfos.Add(Md5JointInfo().SetValue(joint,this.Animation.JointInfos.Count))  
                | StartsWith "bounds" true ->
                    let mutable isbound = true 
                    while isbound  do
                        let bound = animFile.ReadLine()
                        isbound <- not (bound.Contains("}"))
                        if isbound then 
                            let data = getLineData bound
                            let a,b,c,x,y,z = getFloat data.[0],getFloat data.[1],getFloat data.[2],getFloat data.[3],getFloat data.[4],getFloat data.[5]
                            this.Animation.Bounds.Add(Vector3(a,b,c),Vector3(x,y,z))               
                | StartsWith "baseframe" true ->
                    let mutable isFrame = true 
                    let mf = this.Animation.BaseFrame       
                    while isFrame  do
                        let frameLine = animFile.ReadLine()
                        isFrame <- not (frameLine.Contains("}"))
                        if isFrame then mf.SetValue(frameLine)                                           
                | StartsWith "frame" true ->
                    let mutable isFrame = true 
                    let mf = Md5Frame()
                    mf.Index <- getInt (getLineData str).[1]
                    this.Animation.Frames.Add(mf)
                    while isFrame  do
                        let frameLine = animFile.ReadLine()
                        isFrame <- not (frameLine.Contains("}"))
                        if isFrame then mf.SetValue(frameLine)                     
                | _ -> printfn "%s" str    
            animFile.Close()
            //把骨骼动画中,各节点由父骨骼节点坐标转化成模型坐标
            this.Animation.CreateFrameSkeleton()
            //生成纹理
            this.Meshs.ForEach(fun mesh -> if  mesh.TexID = 0 && File.Exists mesh.ShaderPath then mesh.TexID <- TexTure.Load(mesh.ShaderPath))
        //this.CreateVBO() 
    member this.DrawModel() =
        this.Meshs.ForEach(fun mesh -> mesh.DrawVBO())  
    member this.Render()=
        //生成骨骼节点的信息
        let joints = this.Animation.GetCurrentFrameSkeleton()
        //根据骨骼节点生成顶点.也就是蒙皮
        this.Meshs.ForEach(fun mesh -> 
            mesh.Verts.ForEach(fun vert ->
                vert.Position <- Vector3.Zero
                for i in [vert.WeightStart .. vert.WeightStart + vert.WeightCount - 1] do
                    let weigth = mesh.Weights.[i]
                    let joint = joints.[weigth.JointIndex]
                    vert.Position <- vert.Position + (joint.Position + Vector3.Transform(weigth.Position,joint.Quat)) * weigth.Bias
                )
            ) 
        //顶点绘制
        this.Meshs.ForEach(fun mesh -> mesh.Render())        
    member this.CreateVBO() = 
        //先求得顶点的实际数据
        this.Meshs.ForEach(fun mesh -> 
            mesh.Verts.ForEach(fun vert ->
                for i in [vert.WeightStart .. vert.WeightStart + vert.WeightCount - 1] do
                    let weigth = mesh.Weights.[i]
                    let joint = this.Joints.[weigth.JointIndex]
                    vert.Position <- vert.Position + (joint.Position + Vector3.Transform(weigth.Position,joint.Quat)) * weigth.Bias
                )
            ) 
        this.Meshs.ForEach(fun mesh -> mesh.CreateVBO())        

                
                    